using System;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;


namespace ToumeiHitsuji
{
    public partial class SlapStuff 
    {
        public class JSONStorableParamGroup
        {
            private List<JSONStorableParam> _params = new List<JSONStorableParam>();
            private HashSet<string> _paramNames = new HashSet<string>();

            public void RegisterParam(JSONStorableParam param)
            {
                if (_paramNames.Contains(param.name))
                {
                    throw new ArgumentException($"Can't register two parameters with the same name: {param.name}.", "param");
                }
                _params.Add(param);
                _paramNames.Add(param.name);
            }

            public void RegisterFloat(JSONStorableFloat param)
            {
                RegisterParam(param);
            }
            public void RegisterString(JSONStorableString param)
            {
                RegisterParam(param);
            }
            public void RegisterStringChooser(JSONStorableStringChooser param)
            {
                RegisterParam(param);
            }
            public void RegisterBool(JSONStorableBool param)
            {
                RegisterParam(param);
            }

            public JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
            {
                JSONClass jc = new JSONClass();
                foreach (JSONStorableParam param in _params)
                {
                    param.StoreJSON(jc, includePhysical, includeAppearance, forceStore);
                }
                return jc;
            }

            public void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
            {
                foreach (JSONStorableParam param in _params)
                {
                    param.RestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
                }
            }

            public void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
            {
                foreach (JSONStorableParam param in _params)
                {
                    param.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
                }
            }
        }

        public class JSONStorableStringList : JSONStorableParam
        {
            private List<string> _val = new List<string>();
            public List<string> defaultVal;
            public SetStringListCallback setCallbackFunction;
            public AddStringListCallback addCallbackFunction;
            public RemoveStringListCallback removeCallbackFunction;
            public ClearStringListCallback clearCallbackFunction;

            public delegate void SetStringListCallback(List<string> val);
            public delegate void AddStringListCallback(string val);
            public delegate void RemoveStringListCallback(string val, int index);
            public delegate void ClearStringListCallback();

            public virtual List<string> valNoCallback
            {
                get { return _val; }
                set 
                { 
                    if (value == null)
                    {
                        value = new List<string>();
                    }
                    _val = value;
                }
            }

            public List<string> val
            {
                get { return _val; }
                set 
                {
                    if (value == null)
                    {
                        value = new List<string>();
                    }
                    List<string> previousValue = _val;
                    valNoCallback = value;
                    bool valuesChanged = !previousValue.SequenceEqual(valNoCallback);
                    if (valuesChanged)
                    {
                        setCallbackFunction?.Invoke(_val);
                    }
                }
            }

            public virtual bool AddNoCallback(string value)
            {
                _val.Add(value);
                return true;
            }

            public bool Add(string value)
            {
                if (AddNoCallback(value))
                {
                    addCallbackFunction?.Invoke(value);
                    return true;
                }
                return false;
            }

            public virtual bool RemoveNoCallback(string value)
            {
                return _val.Remove(value);
            }

            public bool Remove(string value)
            {
                int index = valNoCallback.IndexOf(value);
                if (RemoveNoCallback(value))
                {
                    removeCallbackFunction?.Invoke(value, index);
                    return true;
                }
                return false;
            }

            public virtual bool RemoveAtNoCallback(int index)
            {
                if (index >= 0 && index < _val.Count)
                {
                    _val.RemoveAt(index);
                    return true;
                }
                return false;
            }
            
            public bool RemoveAt(int index)
            {
                if (index >= 0 && index < _val.Count)
                {
                    string value = valNoCallback[index];
                    if (RemoveAtNoCallback(index))
                    {
                        removeCallbackFunction?.Invoke(value, index);
                        return true;
                    }
                }
                return false;
            }

            public virtual bool ClearNoCallback()
            {
                if (_val.Count > 0)
                {
                    _val.Clear();
                    return true;
                }
                return false;
            }

            public bool Clear()
            {
                if (ClearNoCallback())
                {
                    clearCallbackFunction?.Invoke();
                    return true;
                }
                return false;
            }

            public JSONStorableStringList(string paramName, List<string> startingValue=null, SetStringListCallback callback=null) : base()
            {
                name = paramName;
                isStorable = true;
                isRestorable = true;
                type = Type.None;
                restoreTime = RestoreTime.Normal; //RestoreTime.Late;
                storeType = StoreType.Full;
                setCallbackFunction = callback;
                addCallbackFunction = null;
                removeCallbackFunction = null;
                clearCallbackFunction = null;
                val = startingValue;
                defaultVal = valNoCallback.ToList();   
            }

            public override void SetDefaultFromCurrent()
            {
                defaultVal = valNoCallback.ToList();
            }

            public override void SetValToDefault()
            {
                val = defaultVal.ToList();
            }

            private void restoreFromJSONInternal(JSONClass jc, bool setMissingToDefault = true)
            {
                if (jc.HasKey(name))
                {
                    List<string> restoredVal = new List<string>();
                    foreach (JSONNode jn in jc[name].AsArray)
                    {
                        restoredVal.Add(jn.Value);
                    }
                    val = restoredVal;    
                }
                else if (setMissingToDefault)
                {
                    val = defaultVal;
                }
            }

            public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
            {
                if (restoreTime == RestoreTime.Late)
                {
                    restoreFromJSONInternal(jc, setMissingToDefault);
                }
            }

            public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
            {
                if (restoreTime == RestoreTime.Normal)
                {
                    restoreFromJSONInternal(jc, setMissingToDefault);
                }
            }

            public override bool StoreJSON(JSONClass jc, bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
            {
                if (jc.HasKey(name))
                {
                    //throw new InvalidOperationException($"Parameter name already present in json: {name}");
                    return false;
                }

                if (forceStore || !_val.SequenceEqual(defaultVal))
                {
                    JSONArray stringArray = new JSONArray();
                    foreach(string s in valNoCallback)
                    {
                        stringArray.Add(new JSONData(s));
                    }
                    
                    jc[name] = stringArray;
                    return true;
                }

                return false;
            }
        }
    }
}